#!/usr/bin/env python3
# abacus
#
# Copyright (C) 2022 Aleksandar Radivojevic
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import ast, code

from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from .console import CodeObj, IConsole

class IConsoleBasic(IConsole):
    def __init__(self):
        super().__init__()

        self._ns = {}
        self._str_transformers = []
        self._ast_transformers = []
        self._event_callbacks = {}
        self.interpreter = code.InteractiveInterpreter(self.user_ns)

        self.set_auto_symbol(True)
        self.set_impl_multi(True)

        self.init_ns()

    @property
    def user_ns(self) -> Dict[str, Any]:
        return self._ns

    @property
    def event_callbacks(self) -> Dict[str, List[Callable]]:
        return self._event_callbacks

    @property
    def ast_transformers(self) -> List[ast.NodeTransformer]:
        return self._ast_transformers

    @property
    def str_transformers(self) -> List[Callable]:
        return self._str_transformers

    @staticmethod
    def console_type() -> str:
        return 'basic'

    def run_interactive(self,
                        code: Union[str, List[ast.AST], CodeObj],
                        transform=True):
        if isinstance(code, str):
            if transform:
                code = self.str_transform(code) # TODO: maybe split by line?

            code = ast.parse(code, filename='<input>', mode='exec')

        if transform:
            self.ast_transform(code)

        (stmt, module) = self.compile_ast(code)

        if module is not None:
            self.interpreter.runcode(module)

        self.interpreter.runcode(stmt)

    def compile_ast(self, node: ast.Module) -> Tuple[CodeObj, Optional[CodeObj]]:
        """Compiles the node and returns the last statement as `ast.Interactive`
        and if there are more than one statements then rest of the module is
        also returned

        Returns last statements, rest of statements"""

        stmt = node.body.pop(-1)
        ast.fix_missing_locations(stmt)
        stmt = compile(ast.Interactive(body=[stmt]), filename='<input>', mode='single')

        module = None
        if len(node.body) >= 1:
            module = compile(node, filename='<input>', mode='exec')

        return stmt, module

def main_basic(*, imitate_user_input=False):
    console = IConsoleBasic()
    print(console.welcome_message())

    while True:
        try:
            x = input(':: ')
        except KeyboardInterrupt:
            break

        if not x:
            break

        # if imitate_user_input:
        #     print(':: ' + x)

        console.run_interactive(x)

if __name__ == '__main__':
    main_basic()



#!/usr/bin/env python3
# abacus
#
# Copyright (C) 2022 Aleksandar Radivojevic
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from IPython.terminal.prompts import Prompts, Token

class BasicPrompt(Prompts):
    def in_prompt_tokens(self, cli=None):
        return [(Token.Prompt, ':: ')]

    def out_prompt_tokens(self):
       return []

    def continuation_prompt_tokens(self, cli=None, width=None):
        return [(Token.PromptNum, '... ')]

def load_ipython_extension(ipy):
    prompt = BasicPrompt(ipy)
    prompt.old = ipy.prompts
    ipy.prompts = prompt

def unload_ipython_extension(ipy):
    if hasattr(ipy.prompts, 'old'):
        ipy.prompts = ipy.prompts.old
    else:
        print('Unable to unload prompt for some reason, maybe it was not loaded?')



#!/usr/bin/env python3
# abacus
#
# Copyright (C) 2022 Aleksandar Radivojevic
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import IPython, sys, ast

from traitlets.config import Config as IPythonConfig
from typing import Any, Callable, Dict, List, Union
from ..console import CodeObj, IConsole
from . import prompt, aliases

class IPythonConsole(IConsole):
    def __init__(self, /, ipy):
        super().__init__()

        self.ipython = ipy

        print(self.welcome_message())

        # TODO: config stuff

        self.set_auto_symbol(True)
        self.set_impl_multi(True)

        self.ipython.extension_manager.load_extension(prompt.__name__)
        self.ipython.extension_manager.load_extension(aliases.__name__)

        self.init_ns()

    @property
    def user_ns(self) -> Dict[str, Any]:
        return self.ipython.user_ns

    @property
    def event_callbacks(self) -> Dict[str, List[Callable]]:
        return self.ipython.events.callbacks

    @property
    def ast_transformers(self) -> List[ast.NodeTransformer]:
        return self.ipython.ast_transformers

    @property
    def str_transformers(self) -> List[Callable]:
        return self.ipython.input_transformers_post

    @staticmethod
    def console_type() -> str:
        return 'ipython'

    # TODO:
    def run_interactive(self, code: Union[str, List[ast.AST], CodeObj]) -> Any:
        if isinstance(code, str):
            self.ipython.run_cell(code)
        else:
            raise NotImplementedError()

    def ast_transform(self, node: ast.AST) -> ast.AST:
        return self.ipython.transform_ast(node)

    def str_transform(self, code: str) -> str:
        return self.ipython.transform_cell(code)

def load_ipython_extension(ipy):
    IPythonConsole(ipy)

def main_ipython():
    """Starting point of IPython version of abacus"""
    cfg = IPythonConfig()
    cfg.TerminalIPythonApp.display_banner = False
    cfg.TerminalIPythonApp.quick = True
    cfg.InteractiveShellApp.pylab_import_all = False

    # load abacus directly so it isn't an extension
    cfg.InteractiveShellApp.exec_lines = [
        'from abacus.ipython import load_ipython_extension',
        'load_ipython_extension(get_ipython())',
        'del load_ipython_extension'
    ]

    IPython.start_ipython(sys.argv[1:], config=cfg)

if __name__ == '__main__':
    main_ipython()



#!/usr/bin/env python3
# abacus
#
# Copyright (C) 2022 Aleksandar Radivojevic
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import ast

from abc import abstractmethod, ABCMeta
from typing import (
    Dict, TypeVar, Callable, Any, List, Mapping, Union
)
from . import __version__, __version_info__

CodeObj = TypeVar('CodeObj')

class ConsoleExtension:
    def __init__(self):
        self._registered = False

    def register(self):
        self._registered = True

        return self

    def unregister(self):
        self._registered = False

    def register_toggle(self, state: bool=None):
        if state is None:
            self.register_toggle(not self.is_registered())
        elif state == True:
            if not self.is_registered():
                self.register()
        else:
            if self.is_registered():
                self.unregister()

    def is_registered(self) -> bool:
        return self._registered

class StringTransformer(metaclass=ABCMeta):
    @abstractmethod
    def transform(self, line: str) -> str:
        pass

    def transform_lines(self, lines: List[str]) -> List[str]:
        result = []
        for line in lines:
            result.append(self.transform(line))

        return result

    def __call__(self, lines: List[str]) -> List[str]:
        return self.transform_lines(lines)

class IConsole(metaclass=ABCMeta):
    def __init__(self):
        # importing here so there is no circular import
        from abacus.transformations.auto_symbol import AutoSymbolTransformer
        from abacus.transformations.implied_multiplication import ImplMulTransformer

        # NOTE: making instances not registering them!
        self.ext_auto_symbol = AutoSymbolTransformer(self)
        self.ext_impl_multi = ImplMulTransformer(self)

    @property
    @abstractmethod
    def user_ns(self) -> Dict[str, Any]:
        pass

    @property
    @abstractmethod
    def event_callbacks(self) -> Dict[str, List[Callable]]:
        pass

    @property
    @abstractmethod
    def ast_transformers(self) -> List[ast.NodeTransformer]:
        pass

    @property
    @abstractmethod
    def str_transformers(self) -> List[Callable]:
        pass

    @staticmethod
    @abstractmethod
    def console_type(cls) -> str:
        """Returns string for which type of console it is"""
        pass

    @abstractmethod
    def run_interactive(self,
                        code: Union[str, List[ast.AST], CodeObj],
                        transform=True,
                        imitate_user_input=False):
        """Evaluates the code inside the namespace after ast and string
        transformations and then triggers `post_execute` event

        `transform`: Should input be transformed, do note that not all
        transformations can be done on all types of input:
            `str`: All transformations are performed
            `ast.AST`: Only AST transformation will be performed
            `CodeObj`: No transformation is done"""
        pass

    def set_impl_multi(self, state: bool):
        self.ext_impl_multi.register_toggle(state)

    def set_auto_symbol(self, state: bool):
        self.ext_auto_symbol.register_toggle(state)

    def welcome_message(self) -> str:
        """Greet message shown on startup of abacus"""

        return f"""\t:: Abacus {__version__} ({self.console_type()}) ::"""

    def push(self, _locals: Mapping[str, Any]) -> 'IConsole':
        """Set locals in the user namespace"""

        self.user_ns.update(_locals)

        return self

    def init_ns(self, _locals: Mapping[str, Any]={}) -> 'IConsole':
        """Initializes the namespace with default values"""

        self.push({
            **_locals,
            '__version__': __version__,
            '__version_info__': __version_info__,
            '__abacus__': self.console_type(),
            'abacus': self,
        })

        self.execute('import sympy')
        self.execute('sympy.init_printing(use_unicode=True)')

        return self

    def execute(self, code: Union[str, List[ast.AST], CodeObj]):
        """Executes the code inside the namespace verbatim"""

        if isinstance(code, ast.AST):
            code = compile(code, filename='<input>', mode='exec')

        exec(code, self.user_ns)

    def evaluate(self, code: Union[str, ast.AST, CodeObj]) -> Any:
        """Evaluates the code inside namespace verbatim and returns result of
        the last statement

        Does not print anything"""

        if isinstance(code, ast.AST):
            code = compile(code, filename='<input>', mode='eval')

        return eval(code, self.user_ns)

    def _register_event(self, event: str, handler):
        self.event_callbacks.setdefault(event, []).append(handler)

    def _unregister_event(self, event: str, handler):
        self.event_callbacks.setdefault(event, []).remove(handler)

    def ast_transform(self, node: ast.AST) -> ast.AST:
        """Performs AST transformation on the node

        WARNING: the input `node` may be modified in the process"""

        i: ast.NodeTransformer
        for i in self.ast_transformers:
            i.visit(node)

        return node

    def str_transform(self, code: Union[str, List[str]]) -> Union[str, List[str]]:
        """Performs string transformation on the code"""

        got_str = False

        if isinstance(code, str):
            code = [code]
            got_str = True

        result = code

        i: Callable[[str], str]
        for i in self.str_transformers:
            result = i(result)

        if got_str:
            return result[0]

        return result



#!/usr/bin/env python3
# abacus
#
# Copyright (C) 2022 Aleksandar Radivojevic
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import ast, sympy

from ..console import ConsoleExtension, IConsole

class AutoSymbolTransformer(ast.NodeTransformer, ConsoleExtension):
    def __init__(self, console: IConsole):
        ConsoleExtension.__init__(self)

        self.console = console
        self.symbols = []

    def register(self):
        # auto symbol needs to be ran before the impl multiplication
        self.console.ast_transformers.insert(0, self)
        self.console._register_event('post_execute', self.post_execute)

        return super().register()

    def unregister(self):
        self.console.ast_transformers.remove(self)
        self.console._unregister_event('post_execute', self.post_execute)

        super().unregister()

    def post_execute(self):
        """Deletes symbols found while parsing AST"""
        for i in self.symbols:
            self.console.execute(f'del {i}')

        # remove old symbols
        self.symbols = []

    def _is_symbol(self, expr: ast.Expr):
        """Checks if expr is a defined symbol in user namespace, used after
        all undefined names are defined as symbols"""
        if isinstance(expr, ast.Name):
            return isinstance(self.console.user_ns.get(expr.id), sympy.Symbol)

        return False

    # TODO: if i disable attributes then auto symbols does not work inside
    # functions that are inside an attribute but otherwise it will cause weird
    # confusing errors when using attribute of undefined var
    # def visit_Attribute(self, node: ast.Attribute):
    #     return node

    def visit_Constant(self, node: ast.Constant):
        # wrap integers in `sympy.Integer`
        if isinstance(node.value, int):
            return ast.Call(
                func=ast.Attribute(
                    value=ast.Name(id='sympy', ctx=ast.Load()),
                    attr='Integer',
                    ctx=ast.Load()
                ),
                args=[node],
                keywords=[]
            )

        return node

    def visit_Name(self, node: ast.Name):
        # there is no need to modify assignments or deletion
        if not isinstance(node.ctx, ast.Load):
            return node

        # create names as symbols if not already created
        if node.id not in self.console.user_ns \
            and node.id not in self.symbols \
            and node.id not in __builtins__:
            self.symbols.append(node.id)
            self.console.execute(f'{node.id} = sympy.Symbol("{node.id}")')

        return node

    def visit_Compare(self, node: ast.Compare):
        # run on children so symbols are made
        self.generic_visit(node)

        # TODO: make it work for any number of comparisons..
        if len(node.ops) == 1 and (self._is_symbol(node.left) or self._is_symbol(node.comparators[0])):
            if isinstance(node.ops[0], ast.Eq):
                return ast.Call(
                    func=ast.Attribute(
                        value=ast.Name(id='sympy', ctx=ast.Load()),
                        attr='Eq',
                        ctx=ast.Load()
                    ),
                    args=[
                        node.left,
                        node.comparators[0],
                    ],
                    keywords=[]
                )
            elif isinstance(node.ops[0], ast.NotEq):
                return ast.Call(
                    func=ast.Attribute(
                        value=ast.Name(id='sympy', ctx=ast.Load()),
                        attr='Ne',
                        ctx=ast.Load()
                    ),
                    args=[
                        node.left,
                        node.comparators[0],
                    ],
                    keywords=[]
                )

        return node



#!/usr/bin/env python3
# abacus
#
# Copyright (C) 2022 Aleksandar Radivojevic
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import ast, token, traceback

from keyword import iskeyword
from tokenize import (
    TokenInfo,
    untokenize as _untokenize,
    generate_tokens as _generate_tokens
)
from io import StringIO
from typing import List
from ..console import ConsoleExtension, IConsole, StringTransformer

class ImplMulTransformer(ast.NodeTransformer, StringTransformer, ConsoleExtension):
    def __init__(self, console: IConsole):
        ConsoleExtension.__init__(self)

        self.console = console

    def register(self):
        # needs to run after auto symbol
        self.console.str_transformers.insert(1, self) # TODO: what happens if there is no transformers?
        self.console.ast_transformers.insert(1, self)

        return super().register()

    def unregister(self):
        self.console.str_transformers.remove(self)
        self.console.ast_transformers.remove(self)

        super().unregister()

    @staticmethod
    def _tokenize(input: str) -> List[TokenInfo]:
        return list(_generate_tokens(StringIO(input).readline))

    @staticmethod
    def _token_good(tok: TokenInfo):
        if tok.type == token.NUMBER:
            return True
        elif tok.type == token.NAME:
            return not iskeyword(tok.string)

        return False

    def transform(self, line: str) -> str:
        tokens = self._tokenize(line)

        count = 0 # number of added tokens
        prev = tokens[0]
        for i, cur in enumerate(list(tokens[1:]), 1):
            if self._token_good(cur):
                if self._token_good(prev) or prev.exact_type == token.RPAR:
                    # TODO: it works but includes whitespace which makes the
                    # output weird sometimes
                    tokens.insert(i + count, (token.OP, ' * '))
                    count += 1

            prev = cur

        return _untokenize(tokens)

    def visit_Call(self, node: ast.Call):
        # turns calls into multiplication if name called is not callable
        try:
            result = None
            try:
                result = self.console.evaluate(ast.Expression(body=node.func))
            except NameError:
                # it does not exist
                pass
            except Exception as ex:
                return node

            # if it's callable then do nothing
            if (
                    # skip functions
                    callable(result) \

                    # there is nothing in the function, or just keywords so skip it
                    or len(node.args) == 0
                ):
                return node

            # if there are multiple 'arguments' in the function call then treat
            # it as a tuple
            # NOTE: keywords are ignored
            if len(node.args) > 1:
                right = ast.Tuple(
                    elts=node.args,
                    ctx=ast.Load()
                )
            else:
                right = node.args[0]

            return ast.BinOp(
                left=node.func,
                op=ast.Mult(),
                right=right
            )
        except Exception:
            print('ImplMulTransformer: Error while parsing AST:\n', traceback.format_exc())
            return node
